/*

 DynFileFmtIO Class for C++
 (c) 2004 Jonathan Bettencourt / Kreative Korporation

 This is the standard code base for manipulating DFF (Dynamic File Format) files.


 This code is under the MIT license.

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in the
 Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <string>
#include <map>
#include <iostream>
#include "binarystream.h"
#include "memoryblock.h"
#include "stringmaniplib.h"
#include "dynfilefmtio.h"
using namespace std;


/*  G L O B A L   F U N C T I O N S  */

string dffFlagBitsToUserString(unsigned short int f)
{
	string fs = "lsppfrhdpfxc1234";
	unsigned short int r = f;
	int x;
	for (x=0; x<16; ++x)
	{
		if ((r % 2) == 0) { fs[x] = '-'; }
		r /= 2;
	}
	return fs;
}


/*  D F F R E C O R D   C L A S S  */

dffrecord::dffrecord(void)
{
	memoryblock m(0);
	datatype = 0;
	setflagint(0);
	idnumber = 0;
	readindatalen = 0;
	symbol = (string)"????????";
	title = (string)"";
	thedata = m;
}

unsigned short int dffrecord::getflagint(void)
{
	unsigned short int r = 0;
	int x;
	for (x=15; x>=0; --x)
	{
		r *= 2;
		if (flagbits[x]) { r++; }
	}
	return r;
}

void dffrecord::setflagint(unsigned short int flagz)
{
	unsigned short int r = flagz;
	int x;
	for (x=0; x<16; ++x)
	{
		flagbits[x] = ((r % 2) != 0);
		r /= 2;
	}
}

bool dffrecord::getlock(void)
{
	return flagbits[0];
}

void dffrecord::lock(void)
{
	flagbits[0] = true;
}

void dffrecord::unlock(void)
{
	flagbits[0] = false;
}

string dffrecord::correctedsymbol(void)
{
	return left((symbol+(string)"        "),8);
}


/*  D Y N F I L E F M T I O   C L A S S  */

/* Constructors */

dynfilefmtio::dynfilefmtio(void)
{
	buffersize = 262144;
	notdff = true;
	datastartptr = 0;
	headerstartptr = 0;
	namespacestartptr = 0;
	numrecords = 0;
	expandorcontractbailed = false;
	mystream.setMotorola();
}

dynfilefmtio::dynfilefmtio(char * filename, bool writeAccess, bool clearFile)
{
	buffersize = 262144;
	notdff = true;
	datastartptr = 0;
	headerstartptr = 0;
	namespacestartptr = 0;
	numrecords = 0;
	expandorcontractbailed = false;
	mystream.setMotorola();
	mystream.open(filename, writeAccess, clearFile);
	init(clearFile);
}

dynfilefmtio::dynfilefmtio(string filename, bool writeAccess, bool clearFile)
{
	buffersize = 262144;
	notdff = true;
	datastartptr = 0;
	headerstartptr = 0;
	namespacestartptr = 0;
	numrecords = 0;
	expandorcontractbailed = false;
	mystream.setMotorola();
	mystream.open(filename, writeAccess, clearFile);
	init(clearFile);
}

void dynfilefmtio::init(bool makedffifnot)
{
	unsigned int i;
	string ether;
	if (! mystream.ok() ) {
		notdff = true;
	} else {
		notdff = false;
		mystream.setPosition(0);
		ether = mystream.readStringValueAsString(4);
		if (ether != "%DF2") {
			if (makedffifnot) {
				mystream.setPosition(0);
				mystream.setLength(8);
				mystream.setPosition(0);
				mystream.writeStringValue(4, "%DF2");
				mystream.writeLong(0);
				headerstartptr = 8;
				namespacestartptr = 8;
				datastartptr = 8;
				numrecords = 0;
			} else {
				notdff = true;
				mystream.close();
			}
		} else {
			mystream.setPosition(4);
			numrecords = mystream.readLong();
			headerstartptr = 8;
			if (numrecords == 0) {
				namespacestartptr = 8;
				datastartptr = 8;
			} else {
				namespacestartptr = (8 + (numrecords * 20));
				mystream.setPosition(namespacestartptr);
				for (i=1; i<=numrecords; ++i) { ether = mystream.readPStringAsString(); }
				datastartptr = mystream.getPosition();
			}
			mystream.setPosition(8);
		}
	}
}

dynfilefmtio::~dynfilefmtio(void)
{
	close();
}

/* Opening & Closing Files */

void dynfilefmtio::create(const string & filename)
{
	mystream.setMotorola();
	mystream.open(filename, true, true);
	init(true);
}

void dynfilefmtio::open(const string & filename)
{
	mystream.setMotorola();
	mystream.open(filename, true, false);
	init(false);
}

void dynfilefmtio::open(const string & filename, bool a, bool b)
{
	mystream.setMotorola();
	mystream.open(filename, a, b);
	init(false);
}

void dynfilefmtio::close(void)
{
	mystream.close();
	notdff = true;
	datastartptr = 0;
	headerstartptr = 0;
	namespacestartptr = 0;
	numrecords = 0;
}

/* Internal Functions */

unsigned long int dynfilefmtio::recordDataStart(unsigned long int i)
{
	unsigned long int ofst = 0;
	unsigned long int x;
	if (i>0)
	{
		for (x=0; x<i; x++)
		{
			mystream.setPosition(headerstartptr + (x*20) + 16);
			ofst += mystream.readLong();
		}
	}
	return datastartptr + ofst;
}

unsigned long int dynfilefmtio::recordHeaderStart(unsigned long int i)
{
	return headerstartptr+(i*20);
}

unsigned long int dynfilefmtio::recordNameStart(unsigned long int i)
{
	unsigned int x;
	string ether;
	mystream.setPosition(namespacestartptr);
	if (i>0)
	{
		for (x=0; x<i; ++x) { ether = mystream.readPStringAsString(); }
	}
	x = mystream.getPosition();
	return x;
}

void dynfilefmtio::contract(unsigned long int contractionPoint, unsigned long int numBytes)
{
	char * buffer;
	unsigned long int ps, fs, nbtm, i, paddr, faddr;
	expandorcontractbailed = false;
	if (buffersize < 1024) { buffersize = 262144; }
	ps = mystream.getLength();
	fs = ps-numBytes;
	nbtm = ps-contractionPoint-numBytes;
	for (i=0; i<nbtm; i+=buffersize)
	{
		paddr = contractionPoint + i + numBytes;
		faddr = paddr-numBytes;
		mystream.setPosition(paddr);
		buffer = mystream.read(buffersize);
		mystream.setPosition(faddr);
		mystream.write(buffer, buffersize);
		delete [] buffer;
	}
	mystream.setPosition(contractionPoint);
	mystream.setLength(fs);
}

void dynfilefmtio::expand(unsigned long int expansionPoint, unsigned long int numBytes)
{
	char * buffer;
	unsigned long int ps, fs, nbtm, i, paddr, faddr;
	expandorcontractbailed = false;
	if (buffersize < 1024) { buffersize = 262144; }
	ps = mystream.getLength();
	fs = ps+numBytes;
	nbtm = ps-expansionPoint;
	mystream.setLength(fs);
	for (i=(nbtm / buffersize); i != UINT32_NEGATIVE_1; --i)
	{
		paddr = expansionPoint + (i*buffersize);
		faddr = paddr+numBytes;
		mystream.setPosition(paddr);
		buffer = mystream.read(buffersize);
		mystream.setPosition(faddr);
		mystream.write(buffer, buffersize);
		delete [] buffer;
	}
	mystream.setPosition(expansionPoint);
	mystream.setLength(fs);
}

/* ID-to-Index */

unsigned long int dynfilefmtio::getRecordIndex(const string & symbol, long int idnumber)
{
	unsigned long int i;
	string csymb;
	long int cid;
	long int xx;
	if (numrecords < 1) { return DFF_INDEX_UNDEFINED; }
	for (i=0; i<numrecords; ++i)
	{
		mystream.setPosition(headerstartptr + (i*20));
		csymb = mystream.readStringValueAsString(8);
		cid = mystream.readUShort();
		xx = mystream.readLong();
		xx = mystream.readUShort();
		cid = (xx << 16) + cid;
		if ( (csymb == symbol) && (cid == idnumber) ) { return i; }
	}
	return DFF_INDEX_UNDEFINED;
}

unsigned long int dynfilefmtio::getRecordIndexByName(const string & symbol, const string & name)
{
	unsigned long int i;
	string csymb;
	string cname;
	if (numrecords < 1) { return DFF_INDEX_UNDEFINED; }
	for (i=0; i<numrecords; ++i)
	{
		mystream.setPosition(headerstartptr + (i*20));
		csymb = mystream.readStringValueAsString(8);
		mystream.setPosition(recordNameStart(i));
		cname = mystream.readPStringAsCString();
		if ( (csymb == symbol) && (cname == name) ) { return i; }
	}
	return DFF_INDEX_UNDEFINED;
}

/* Traversal Functions */

unsigned long int dynfilefmtio::maxIndexValue(void)
{
	return (numrecords-1);
}

unsigned long int dynfilefmtio::numSymbols(void)
{
	map<string,bool> defdSymbols;
	unsigned long int i,nn;
	string csymb;
	nn = 0;
	if (numrecords < 1) { return 0; }
	for (i = 0; i<numrecords; i++)
	{
		mystream.setPosition(headerstartptr+(i*20));
		csymb = mystream.readStringValueAsString(8);
		if ( defdSymbols.count(csymb) == 0 )
		{
			nn++;
			defdSymbols.insert( pair<string,bool>(csymb,true) );
		}
	}
	return nn;
}

string dynfilefmtio::nthSymbol(unsigned long int n)
{
	map<string,bool> defdSymbols;
	unsigned long int i,nn;
	string csymb;
	nn = 0;
	if (numrecords < 1) { return (string)""; }
	for (i = 0; i<numrecords; i++)
	{
		mystream.setPosition(headerstartptr+(i*20));
		csymb = mystream.readStringValueAsString(8);
		if ( defdSymbols.count(csymb) == 0 )
		{
			nn++;
			if (nn == n) { return csymb; }
			defdSymbols.insert( pair<string,bool>(csymb,true) );
		}
	}
	return (string)"";
}

unsigned long int dynfilefmtio::numNumbers(const string & theSymbol)
{
	unsigned long int i,nn;
	string csymb;
	nn = 0;
	if (numrecords < 1) { return 0; }
	for (i = 0; i<numrecords; i++)
	{
		mystream.setPosition(headerstartptr+(i*20));
		csymb = mystream.readStringValueAsString(8);
		if ( csymb == theSymbol )
		{
			nn++;
		}
	}
	return nn;
}

int dynfilefmtio::nthNumber(const string & theSymbol, unsigned long int n)
{
	unsigned long int i, nn;
	int cid, xx;
	string csymb;
	nn = 0;
	if (numrecords < 1) { return 0; }
	for (i = 0; i<numrecords; i++)
	{
		mystream.setPosition(headerstartptr+(i*20));
		csymb = mystream.readStringValueAsString(8);
		cid = mystream.readUShort();
		xx = mystream.readLong();
		xx = mystream.readUShort();
		cid = (xx << 16) + cid;
		if ( csymb == theSymbol )
		{
			nn++;
			if (nn == n) { return cid; }
		}
	}
	return 0;
}

long int dynfilefmtio::nextAvailableNumber(const string & theSymbol)
{
	return nextAvailableNumber(theSymbol, 0);
}

long int dynfilefmtio::nextAvailableNumber(const string & theSymbol, long int startNum)
{
	map<long int,bool> foundNumbers;
	unsigned long int i;
	long int cid, xx;
	string csymb;
	long int n = startNum;
	if (numrecords < 1) { return 0; }
	for (i = 0; i<numrecords; i++)
	{
		mystream.setPosition(headerstartptr+(i*20));
		csymb = mystream.readStringValueAsString(8);
		cid = mystream.readUShort();
		xx = mystream.readLong();
		xx = mystream.readUShort();
		cid = (xx << 16) + cid;
		if (csymb == theSymbol) { foundNumbers.insert( pair<long int,bool>(cid,true) ); }
	}
	while ( foundNumbers.count(n) > 0 ) { n++; }
	return n;
}

/* Misc Functions */

void dynfilefmtio::addRecord(dffrecord theRcd)
{
	string c;
	expand(datastartptr, theRcd.title.size()+1);
	mystream.setPosition(datastartptr);
	c = theRcd.title;
	mystream.writePStringAsString(c);
	datastartptr += theRcd.title.size()+1;
	expand(namespacestartptr,20);
	mystream.setPosition(namespacestartptr);
	mystream.write(theRcd.correctedsymbol().c_str(),8);
	mystream.writeShort(theRcd.idnumber % 65536);
	mystream.writeShort(theRcd.datatype);
	mystream.writeShort(theRcd.getflagint());
	mystream.writeShort(theRcd.idnumber / 65536);
	mystream.writeLong(theRcd.thedata.getSize());
	namespacestartptr += 20;
	datastartptr += 20;
	if ( theRcd.thedata.getSize() != 0 )
	{
		mystream.setPosition(mystream.getLength());
		char * ch;
		ch = theRcd.thedata.getStringValue(0,theRcd.thedata.getSize());
		mystream.write(ch, theRcd.thedata.getSize());
		delete [] ch;
	}
	numrecords++;
	mystream.setPosition(4);
	mystream.writeLong(numrecords);
}

void dynfilefmtio::deleteAllRecords(void)
{
	mystream.setLength(8);
	mystream.setPosition(4);
	mystream.writeLong(0);
	datastartptr = 8;
	headerstartptr = 8;
	namespacestartptr = 8;
	numrecords = 0;
}

/* Delete/Get/Set By Index */

void dynfilefmtio::deleteRecordByIndex(unsigned long int theIndex)
{
	unsigned long int nLen, dLen, nStrt, dStrt;
	if (! (theIndex > maxIndexValue())) {
		nStrt = recordNameStart(theIndex);
		mystream.setPosition(nStrt);
		nLen = mystream.readByte()+1;
		mystream.setPosition(recordHeaderStart(theIndex)+16);
		dLen = mystream.readLong();
		dStrt = recordDataStart(theIndex);
		contract(dStrt, dLen);
		contract(nStrt, nLen);
		contract(recordHeaderStart(theIndex), 20);
		namespacestartptr -= 20;
		datastartptr -= nLen;
		datastartptr -= 20;
		numrecords--;
		mystream.setPosition(4);
		mystream.writeLong(numrecords);
	}
}

dffrecord dynfilefmtio::getRecordByIndex(unsigned long int theIndex)
{
	long int xx;
	unsigned long int xxx;
	dffrecord r;
	memoryblock m;
	if (! (theIndex > maxIndexValue())) {
		mystream.setPosition(recordHeaderStart(theIndex));
		r.symbol = mystream.readStringValueAsCString(8);
		r.idnumber = mystream.readUShort();
		r.datatype = mystream.readShort();
		r.setflagint(mystream.readUShort());
		xx = mystream.readUShort();
		r.idnumber = (xx << 16) + r.idnumber;
		xxx = mystream.readLong();
		r.readindatalen = xxx;
		mystream.setPosition(recordNameStart(theIndex));
		r.title = mystream.readPStringAsCString();
		mystream.setPosition(recordDataStart(theIndex));
		m.setSize(xxx);
		if ( xxx != 0 )
		{
			char * ch;
			ch = mystream.read(xxx);
			m.setStringValue(0, xxx, ch);
			delete [] ch;
		}
		r.thedata = m;
	}
	return r;
}

memoryblock dynfilefmtio::getRecordDataByIndex(unsigned long int theIndex)
{
	unsigned long int xx;
	memoryblock m(0);
	if (! (theIndex > maxIndexValue())) {
		mystream.setPosition(recordHeaderStart(theIndex)+16);
		xx = mystream.readLong();
		mystream.setPosition(recordDataStart(theIndex));
		m.setSize(xx);
		if ( xx != 0 )
		{
			char * ch;
			ch = mystream.read(xx);
			m.setStringValue(0, xx, ch);
			delete [] ch;
		}
	}
	return m;
}

int dynfilefmtio::getRecordDataTypeByIndex(unsigned long int theIndex)
{
	if (theIndex > maxIndexValue()) { return 0; }
	mystream.setPosition(recordHeaderStart(theIndex) + 10);
	return mystream.readShort();
}

unsigned long int dynfilefmtio::getRecordLengthByIndex(unsigned long int theIndex)
{
	if (theIndex > maxIndexValue()) { return 0; }
	mystream.setPosition(recordHeaderStart(theIndex) + 16);
	return mystream.readLong();
}

unsigned short int dynfilefmtio::getRecordFlagsByIndex(unsigned long int theIndex)
{
	if (theIndex > maxIndexValue()) { return 0; }
	mystream.setPosition(recordHeaderStart(theIndex) + 12);
	return mystream.readUShort();
}

string dynfilefmtio::getRecordNameByIndex(unsigned long int theIndex)
{
	if (theIndex > maxIndexValue()) { return (string)""; }
	mystream.setPosition(recordNameStart(theIndex));
	return mystream.readPStringAsCString();
}

long int dynfilefmtio::getRecordNumberByIndex(unsigned long int theIndex)
{
	long int xx;
	if (theIndex > maxIndexValue()) { return 0; }
	mystream.setPosition(recordHeaderStart(theIndex) + 8);
	xx = mystream.readShort();
	mystream.setPosition(recordHeaderStart(theIndex) + 14);
	xx += (mystream.readShort() << 16);
	return xx;
}

string dynfilefmtio::getRecordSymbolByIndex(unsigned long int theIndex)
{
	if (theIndex > maxIndexValue()) { return 0; }
	mystream.setPosition(recordHeaderStart(theIndex));
	return mystream.readStringValueAsString(8);
}

void dynfilefmtio::setRecordByIndex(unsigned long int theIndex, dffrecord & r)
{
	if (theIndex > maxIndexValue()) { throw "Attempt to write beyond last record."; }
	memoryblock m = r.thedata;
	mystream.setPosition(recordHeaderStart(theIndex));
	mystream.write(r.correctedsymbol().c_str(),8);
	mystream.writeUShort(r.idnumber % 65536);
	mystream.writeShort(r.datatype);
	mystream.writeUShort(r.getflagint());
	mystream.writeShort(r.idnumber / 65536);
	mystream.writeLong(m.getSize());
	setRecordDataByIndex(theIndex, m);
	setRecordNameByIndex(theIndex, r.title);
}

void dynfilefmtio::setRecordDataByIndex(unsigned long int theIndex, memoryblock & m)
{
	if (theIndex > maxIndexValue()) { throw "Attempt to write beyond last record."; }
	unsigned long int ds;
	unsigned long int oldDLen, newDLen;
	ds = recordDataStart(theIndex);
	mystream.setPosition(recordHeaderStart(theIndex) + 16);
	oldDLen = mystream.readLong();
	newDLen = m.getSize();
	if ( newDLen > oldDLen )
	{
		expand( ds+oldDLen, diff(newDLen,oldDLen) );
		if ( expandorcontractbailed ) { throw "Failed binarystream expansion."; }
	}
	else if ( newDLen < oldDLen )
	{
		contract( ds+newDLen, diff(newDLen,oldDLen) );
		if ( expandorcontractbailed ) { throw "Failed binarystream contraction."; }
	}
	if ( newDLen != 0 )
	{
		mystream.setPosition(ds);
		char * ch;
		ch = m.getStringValue(0,newDLen);
		mystream.write(ch,newDLen);
		delete [] ch;
	}
	mystream.setPosition(recordHeaderStart(theIndex) + 16);
	mystream.writeLong(newDLen);
}

void dynfilefmtio::setRecordDataTypeByIndex(unsigned long int theIndex, int dt)
{
	if (theIndex > maxIndexValue()) { throw "Attempt to write beyond last record."; }
	mystream.setPosition(recordHeaderStart(theIndex) + 10);
	mystream.writeShort(dt);
}

void dynfilefmtio::setRecordFlagsByIndex(unsigned long int theIndex, unsigned short int f)
{
	if (theIndex > maxIndexValue()) { throw "Attempt to write beyond last record."; }
	mystream.setPosition(recordHeaderStart(theIndex) + 12);
	mystream.writeUShort(f);
}

void dynfilefmtio::setRecordNameByIndex(unsigned long int theIndex, const string & nm)
{
	if (theIndex > maxIndexValue()) { throw "Attempt to write beyond last record."; }
	unsigned long int ns;
	int oldNLen, newNLen;
	ns = recordNameStart(theIndex);
	mystream.setPosition(ns);
	oldNLen = mystream.readByte() + 1;
	newNLen = nm.size() + 1;
	if ( newNLen > oldNLen )
	{
		expand( ns+oldNLen, diff(newNLen,oldNLen) );
		if ( expandorcontractbailed ) { throw "Failed binarystream expansion."; }
		datastartptr += diff(newNLen,oldNLen);
	}
	else if ( newNLen < oldNLen )
	{
		contract( ns+newNLen, diff(newNLen,oldNLen) );
		if ( expandorcontractbailed ) { throw "Failed binarystream contraction."; }
		datastartptr -= diff(newNLen,oldNLen);
	}
	mystream.setPosition(ns);
	mystream.writePStringAsString(nm);
}

void dynfilefmtio::setRecordNumberByIndex(unsigned long int theIndex, long int n)
{
	if (theIndex > maxIndexValue()) { throw "Attempt to write beyond last record."; }
	mystream.setPosition(recordHeaderStart(theIndex) + 8);
	mystream.writeUShort( n % 65536 );
	mystream.setPosition(recordHeaderStart(theIndex) + 14);
	mystream.writeShort( n / 65536 );
}

void dynfilefmtio::setRecordSymbolByIndex(unsigned long int theIndex, const string & s)
{
	if (theIndex > maxIndexValue()) { throw "Attempt to write beyond last record."; }
	mystream.setPosition(recordHeaderStart(theIndex));
	mystream.writeStringValueAsString(8, s);
}

/* Delete/Get/Set By ID */

void dynfilefmtio::deleteRecord(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx != DFF_INDEX_UNDEFINED )
	{
		deleteRecordByIndex(idx);
	}
}

dffrecord dynfilefmtio::getRecord(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { dffrecord r; return r; }
	return getRecordByIndex(idx);
}

memoryblock dynfilefmtio::getRecordData(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { memoryblock m; return m; }
	return getRecordDataByIndex(idx);
}

int dynfilefmtio::getRecordDataType(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordDataTypeByIndex(idx);
}

unsigned long int dynfilefmtio::getRecordLength(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordLengthByIndex(idx);
}

unsigned short int dynfilefmtio::getRecordFlags(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordFlagsByIndex(idx);
}

string dynfilefmtio::getRecordName(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { return (string)""; }
	return getRecordNameByIndex(idx);
}

long int dynfilefmtio::getRecordNumber(const string & symb, long int id)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordNumberByIndex(idx);
}

void dynfilefmtio::setRecord(const string & symb, long int id, dffrecord & r)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { addRecord(r); }
	else { setRecordByIndex(idx, r); }
}

void dynfilefmtio::setRecordData(const string & symb, long int id, memoryblock & m)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED )
	{
		dffrecord r;
		r.symbol = symb;
		r.idnumber = id;
		r.thedata = m;
		addRecord(r);
	}
	else { setRecordDataByIndex(idx, m); }
}

void dynfilefmtio::setRecordDataType(const string & symb, long int id, int dt)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordDataTypeByIndex(idx, dt);
}

void dynfilefmtio::setRecordFlags(const string & symb, long int id, unsigned short int f)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordFlagsByIndex(idx, f);
}

void dynfilefmtio::setRecordName(const string & symb, long int id, const string & nm)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordNameByIndex(idx, nm);
}

void dynfilefmtio::setRecordNumber(const string & symb, long int id, long int n)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordNumberByIndex(idx, n);
}

void dynfilefmtio::setRecordSymbol(const string & symb, long int id, const string & ns)
{
	unsigned long int idx = getRecordIndex(symb, id);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordSymbolByIndex(idx, ns);
}

/* Delete/Get/Set By Name */

void dynfilefmtio::deleteRecordByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx != DFF_INDEX_UNDEFINED )
	{
		deleteRecordByIndex(idx);
	}
}

dffrecord dynfilefmtio::getRecordByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { dffrecord r; return r; }
	return getRecordByIndex(idx);
}

memoryblock dynfilefmtio::getRecordDataByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { memoryblock m; return m; }
	return getRecordDataByIndex(idx);
}

int dynfilefmtio::getRecordDataTypeByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordDataTypeByIndex(idx);
}

unsigned long int dynfilefmtio::getRecordLengthByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordLengthByIndex(idx);
}

unsigned short int dynfilefmtio::getRecordFlagsByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordFlagsByIndex(idx);
}

string dynfilefmtio::getRecordNameByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { return (string)""; }
	return getRecordNameByIndex(idx);
}

long int dynfilefmtio::getRecordNumberByName(const string & symb, const string & title)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { return 0; }
	return getRecordNumberByIndex(idx);
}

void dynfilefmtio::setRecordByName(const string & symb, const string & title, dffrecord & r)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { addRecord(r); }
	else { setRecordByIndex(idx, r); }
}

void dynfilefmtio::setRecordDataByName(const string & symb, const string & title, memoryblock & m)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED )
	{
		dffrecord r;
		r.symbol = symb;
		r.idnumber = nextAvailableNumber(symb);
		r.title = title;
		r.thedata = m;
		addRecord(r);
	}
	else { setRecordDataByIndex(idx, m); }
}

void dynfilefmtio::setRecordDataTypeByName(const string & symb, const string & title, int dt)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordDataTypeByIndex(idx, dt);
}

void dynfilefmtio::setRecordFlagsByName(const string & symb, const string & title, unsigned short int f)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordFlagsByIndex(idx, f);
}

void dynfilefmtio::setRecordNameByName(const string & symb, const string & title, const string & nm)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordNameByIndex(idx, nm);
}

void dynfilefmtio::setRecordNumberByName(const string & symb, const string & title, long int n)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordNumberByIndex(idx, n);
}

void dynfilefmtio::setRecordSymbolByName(const string & symb, const string & title, const string & ns)
{
	unsigned long int idx = getRecordIndexByName(symb, title);
	if ( idx == DFF_INDEX_UNDEFINED ) { throw "Attempt to change nonexistent record."; }
	setRecordSymbolByIndex(idx, ns);
}